from io import BytesIO
import os, json


def _app_json_parse(body: str):
    return json.loads(body)


def _multipart_form_parse(body: str, headers: dict):
    from multipart import MultipartParser, to_bytes, parse_options_header

    _, options = parse_options_header(headers["Content-Type"])
    boundary = options["boundary"].encode("utf-8")
    p = MultipartParser(BytesIO(to_bytes(body)), boundary)  # TODO: direct from fs
    return [{"name": i.name, "value": i.value} for i in p]


def get_raw_request(local_file="request.json"):
    try:
        with open(os.getenv("REQUEST_FILE", local_file), encoding="utf-8") as f:
            request = json.loads(f.read())
            body = request["body"]
            query = request["query"]
            headers = request["headers"]
        return body, query, headers

    except Exception:
        return None, None, None


def get_request(local_file="request.json"):
    body, query, headers = get_raw_request(local_file)
    try:
        if "application/json" in headers["Content-Type"]:
            return _app_json_parse(body), query, headers
        elif "multipart/form-data" in headers["Content-Type"]:
            return _multipart_form_parse(body, headers), query, headers
        else:
            return body, query, headers
    except Exception:
        return body, query, headers


def send_response(body="", status_code=200, headers={}, local_file="response.json"):
    response = {"status_code": status_code, "body": body, "headers": headers}
    path = os.getenv("RESPONSE_FILE", local_file)
    content = json.dumps(response)
    with open(path, "w", encoding="utf-8") as f:
        f.write(content)


def send_json(data={}, status_code=200, headers={}, local_file="response.json"):
    send_response(
        body=json.dumps(data, allow_nan=False),
        headers={**headers, "content-type": "application/json"},
        status_code=status_code,
        local_file=local_file,
    )



import unittest
from pathlib import Path
from .utils import assert_form
import json
import os


class TestRunSteps(unittest.TestCase):
    def setUp(self) -> None:
        os.chdir(Path(__file__).parent)

    def test_back_and_forth(self):
        code_path = "./resources/test_run_steps/test_back_and_forth/code.py"
        msgs_path = "./resources/test_run_steps/test_back_and_forth/messages.json"
        code = open(code_path).read()
        msgs = json.load(open(msgs_path))
        assert_form(code, msgs, session_id="session-id")



from abstra_server.runtimes.forms.message_handler import MessageBroker
from abstra_server.contract import forms_contract
from abstra_server.session import LiveSession
from concurrent.futures import ThreadPoolExecutor
from collections import deque
from json import loads, dumps
import queue, typing
from abstra_server.overloads import __overload_abstra_forms_sdk
from abstra_server.utils import formated_traceback_error_message

__overload_abstra_forms_sdk()


class MockConnection:
    msgs: queue.Queue

    def __init__(self, browser_msgs) -> None:
        self.browser_msgs = deque(browser_msgs)
        self.msgs = queue.Queue()

    def send(self, data: str):
        self.msgs.put(["python", loads(data)])

    def receive(self) -> str:
        if len(self.browser_msgs) == 0:
            self.close()

        self.msgs.put(["browser", self.browser_msgs[0]])
        return dumps(self.browser_msgs.popleft())

    def close(self, error_msg: typing.Optional[str] = None):
        self.msgs.put(["close", error_msg])

    @property
    def connected(self):
        raise NotImplementedError()


def assert_form_sync(conn: MockConnection, code: str, session_id: str):
    try:
        error_msg = None

        session = LiveSession(conn, "forms", "path")
        session.id = session_id
        broker = MessageBroker(session)
        session.context.update(broker=broker)
        session.send(forms_contract.SessionIdMessage(session.id))
        namespace: typing.Dict[str, typing.Any] = {}

        exec(code, namespace, namespace)
    except Exception as e:
        print(formated_traceback_error_message(e))

        error_msg = str(e)
        raise e
    finally:
        conn.close(error_msg)


def assert_form(code: str, msg_list: list, session_id: str):
    msgs: typing.Deque[list] = deque(msg_list)
    executor = ThreadPoolExecutor()
    browser_msgs = [msg[1] for msg in msgs if msg[0] == "browser"]
    conn = MockConnection(browser_msgs)
    executor.submit(assert_form_sync, conn, code, session_id)

    while True:
        py_msg = conn.msgs.get(block=True, timeout=2)
        if py_msg[0] == "close":
            if py_msg[1] is None:
                break
            else:
                raise Exception(py_msg[1])
        if len(msgs) == 0:
            raise Exception(f"Code is sending unexpected messages: {py_msg}")
        elif msgs[0] != py_msg:
            raise Exception(f"Got {py_msg}, next expected messages: {msgs}")
        else:
            msgs.popleft()



import json

from typing import Dict


def deserialize(string: str) -> Dict:
    return json.loads(string)


def serialize(obj: Dict) -> bytes:
    return bytes(json.dumps(obj, allow_nan=False), "UTF-8")


def get_single_value(answer: Dict):
    return list(answer.values())[0]



import flask_sock, simple_websocket

from ...session import LiveSession
from .program import PythonProgram
from ...api.classes import DashJSON
from ...contract import dashes_contract
from .message_handler import MessageHandler


def run_dash(conn: flask_sock.Server, dash: DashJSON, code: str):
    session = LiveSession(conn, "dashes", dash.path)
    py = PythonProgram(dash, code)
    msg_handler = MessageHandler(py, session, main_file=dash.file)
    session.send(dashes_contract.ExecutionIdMessage(session.id))
    while True:
        try:
            type, data = session.recv()
            msg_handler.handle(type, data)
        except simple_websocket.ConnectionError:
            print("connection closed")



import unittest
from pathlib import Path
from abstra_cli.utils import remove_filepath_prefix
from .fixtures import init_dir, clear_dir
from abstra.widgets.file_utils import convert_file


class TestFileUtils(unittest.TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self):
        clear_dir(self.root)

    def test_remove_filepath_prefix_no_dot(self):
        filepath = "folder/file"
        prefix = "folder"
        self.assertEqual(remove_filepath_prefix(filepath, prefix), "file")

    def test_remove_filepath_prefix_dotted_both(self):
        filepath = "./folder/file"
        prefix = "./folder"
        self.assertEqual(remove_filepath_prefix(filepath, prefix), "file")

    def test_remove_filepath_prefix_dotted_filepath(self):
        filepath = "./folder/file"
        prefix = "folder"
        self.assertEqual(remove_filepath_prefix(filepath, prefix), "file")

    def test_remove_filepath_prefix_dotted_prefix(self):
        filepath = "folder/file"
        prefix = "./folder"
        self.assertEqual(remove_filepath_prefix(filepath, prefix), "file")

    def test_convert_file_preserve_file_name(self):
        path = self.root.joinpath("file.txt").absolute()
        path.write_text("test")
        path_str = str(path)
        converted = convert_file(path_str)
        self.assertTrue(converted.endswith("file.txt"))



import tempfile
import unittest
from uuid import uuid4 as uuid
from pathlib import Path

from abstra_server.api import API

from .fixtures import init_dir, clear_dir
from pathlib import Path


class TestWorkspace(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_start_at_relative_new_dir(self):
        root = Path("./relative-non-new-dir")
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_start_at_relative_new_deep_dir(self):
        root = Path("./relative-non-new-dir/deep/dir")
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_start_at_absolute_new_dir(self):
        root = Path(tempfile.gettempdir()).joinpath("absolute-non-new-dir").absolute()
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_start_at_absolute_new_deep_dir(self):
        root = (
            Path(tempfile.gettempdir())
            .joinpath("absolute-non-new-dir/deep/dir")
            .absolute()
        )
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_start_at_relative_new_deep_dir_ending_with_slash(self):
        root = Path("./relative-non-new-dir/deep/dir/")
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_start_at_absolute_new_deep_dir_ending_with_slash(self):
        root = (
            Path(tempfile.gettempdir())
            .joinpath("absolute-non-new-dir/deep/dir/")
            .absolute()
        )
        try:
            api = API(root=root)
            forms = api.get_forms()
            self.assertEqual(forms, [])
            api.create_form()
        finally:
            clear_dir(root)

    def test_api_update(self):
        api = API(root=self.root)

        api.update_workspace({"name": "test-workspace-updated"})

        workspace = api.get_workspace()

        self.assertEqual(workspace.name, "test-workspace-updated")



import typing


class Reactive:
    type = "reactive"

    def __init__(self, callback: typing.Callable):
        self.callback = callback
        self.page = None

    def json(self, payload):
        self.page = self.callback(payload)
        return self.page.json(payload) if self.page else []

    def get_widgets(self):
        return self.page.get_concrete_widgets() if self.page else []



import logging, os, fire, pathlib
from abstra_server.apps import serve

from . import messages
from .deploy import deploy
from .utils.server import get_free_port
from .version import check_latest_version

PORT = os.getenv("PORT")


class CLI(object):
    """
    A CLI to manage your Abstra Cloud Workspace environment.

    usage: abstra <command> <resource> [<argument> ...] [parameters]
    """

    def deploy(self, workspace_root: str = "."):
        deploy(workspace_root=pathlib.Path(workspace_root))

    def version(self):
        check_latest_version()

    def serve(
        self,
        workspace_root: str = ".",
        port: int = 3000,
        debug=False,
        use_reloader=False,
        load_dotenv=True,
    ):
        default_port = int(PORT or port)
        free_port = get_free_port(default_port=default_port)

        if not debug:
            log = logging.getLogger("werkzeug")
            log.setLevel(logging.WARNING)

        messages.serve_message(free_port)
        check_latest_version()

        serve(
            workspace_root=pathlib.Path(workspace_root),
            port=free_port,
            debug=debug,
            use_reloader=use_reloader,
            load_dotenv=load_dotenv,
        )


def _SeparateFlagArgs(args):
    try:
        index = args.index("--help")
        args = args[:index]
        return args, ["--help"]
    except ValueError:
        return args, []


def main():
    fire.core.parser.SeparateFlagArgs = _SeparateFlagArgs
    fire.Fire(CLI)


if __name__ == "__main__":
    main()



import os
from pathlib import Path

from .utils import CREDENTIALS_FILE
from .path_resolver import resolve_cwd


def get_credentials(root_path: Path):
    if os.getenv("ABSTRA_API_TOKEN"):
        return os.getenv("ABSTRA_API_TOKEN")

    credentials_path = Path(root_path, CREDENTIALS_FILE)
    if not credentials_path.exists():
        return None

    return credentials_path.read_text(encoding="utf-8").strip()


def delete_credentials(root_path: Path):
    credentials_path = Path(root_path, CREDENTIALS_FILE)

    if credentials_path.exists():
        credentials_path.unlink()


def set_credentials(root_path: Path, token: str):
    credentials_path = Path(root_path, CREDENTIALS_FILE)
    credentials_path.parent.mkdir(exist_ok=True)

    credentials_path.write_text(token, encoding="utf-8")


def resolve_headers(root: Path = Path(".")):
    root_path = resolve_cwd(root)
    credentials = get_credentials(root_path)

    if not credentials:
        return None

    return {"Api-Authorization": f"Bearer {credentials}"}



import os, tempfile, urllib.request, zipfile, requests, uuid, pathlib

from .path_resolver import resolve_cwd
from .credentials import resolve_headers
from .utils.file import files_from_directory

CLOUD_API_ENDPOINT = (
    os.environ.get("CLOUD_API_ENDPOINT") or "https://cloud-api.abstra.cloud"
)


def _generate_zip_file(root_path: pathlib.Path) -> pathlib.Path:
    zip_path = pathlib.Path(tempfile.gettempdir(), f"{uuid.uuid4()}.zip")
    with zipfile.ZipFile(zip_path, "w") as zip_file:
        for file in files_from_directory(root_path):
            zip_file.write(file, file.relative_to(root_path))
    return zip_path


def _create_build(headers: dict) -> dict:
    return requests.post(f"{CLOUD_API_ENDPOINT}/cli/builds", headers=headers).json()


def _upload_file(url: str, file_path: pathlib.Path, headers: dict):
    with file_path.open("rb") as f:
        req = urllib.request.Request(url=url, method="PUT", data=f.read())
        urllib.request.urlopen(req)


def _update_build(build_id: str, headers: dict):
    requests.patch(
        f"{CLOUD_API_ENDPOINT}/cli/builds/{build_id}",
        headers=headers,
    )


def deploy(workspace_root: pathlib.Path = pathlib.Path(".")):
    root_path = resolve_cwd(workspace_root)

    headers = resolve_headers(root=root_path)
    if not headers:
        print("No API token found. Please login with `abstra login`")
        return

    data = _create_build(headers)
    url = data["url"]
    build_id = data["buildId"]
    zip_path = _generate_zip_file(root_path)
    _upload_file(url=url, file_path=zip_path, headers=headers)

    _update_build(headers=headers, build_id=build_id)



from ..widget_base import Output
from typing import Union
from ..file_utils import convert_file
import io


class FileOutput(Output):
    type = "file-output"

    def __init__(self, file: Union[str, io.IOBase], **kwargs):
        self.file = file
        self.download_text = kwargs.get("download_text", "Download")
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "fileUrl": convert_file(self.file),
            "downloadText": self.download_text,
            "columns": self.columns,
            "fullWidth": self.full_width,
        }



from abstra.forms.page_response import PageResponse
import unittest
from uuid import uuid4 as uuid
from pathlib import Path

from abstra.tables import TableNotFound, get_db, get_table

from .fixtures import init_dir, clear_dir


class TestTables(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_create_table(self):
        db = get_db()

        created_table = db.create_table()

        retrieved_table = db.get_table(created_table.name)

        self.assertEqual(created_table.name, retrieved_table.name)

    def test_delete_table(self):
        db = get_db()

        created_table = db.create_table()

        db.delete_table(created_table.name)

        with self.assertRaises(TableNotFound):
            db.get_table(created_table.name)

    def test_update_table_name(self):
        db = get_db()

        created_table = db.create_table()

        db.update_table(created_table.name, dict(name="new_name"))

        table = db.get_table("new_name")

        self.assertEqual(table.name, "new_name")

    def test_update_table_columns_insertion(self):
        db = get_db()

        my_table = db.create_table("my_table", [dict(name="my_column")])
        cols = [c.editor_dto for c in my_table.columns]
        cols.append(dict(name="my_column2"))

        db.update_table("my_table", dict(columns=cols))

        cols = get_table("my_table").columns

        self.assertEqual(len(cols), 5)
        self.assertIn("my_column2", [c.name for c in cols])

    def test_update_table_columns_deletion(self):
        db = get_db()

        my_table = db.create_table(
            "my_table", [dict(name="my_column"), dict(name="my_column2")]
        )
        cols = [c.editor_dto for c in my_table.columns]
        cols.pop()

        db.update_table("my_table", dict(columns=cols))

        cols = get_table("my_table").columns

        self.assertEqual(len(cols), 4)
        self.assertNotIn("my_column2", [c.name for c in cols])

    def test_update_table_columns_rename(self):
        db = get_db()

        my_table = db.create_table("my_table", [dict(name="my_column")])
        cols = [c.editor_dto for c in my_table.columns]
        col = [c for c in cols if c["name"] == "my_column"][0]
        col["name"] = "renamed_column"

        db.update_table("my_table", dict(columns=cols))

        cols = get_table("my_table").columns

        self.assertEqual(len(cols), 4)
        self.assertIn("renamed_column", [c.name for c in cols])
        self.assertNotIn("my_column", [c.name for c in cols])

    def test_create_column(self):
        db = get_db()

        table = db.create_table()
        col = db.create_column(table.name)
        cols = db.get_columns(table.name)

        self.assertIn(col.name, [c.name for c in cols])

        db.delete_column(table.name, col.name)

        cols = db.get_columns(table.name)
        self.assertNotIn(col.name, [c.name for c in cols])

    def test_crud_data(self):
        db = get_db()

        table = db.create_table()

        name_column = db.create_column(table.name)
        db.update_column(table.name, name_column.name, dict(name="name"))

        email_column = db.create_column(table.name)
        db.update_column(table.name, email_column.name, dict(name="email"))

        crud_table = get_table(table.name)
        crud_table.insert(
            dict(name="Michael Scott", email="michael.scott@dundermifflin.com")
        )
        crud_table.insert(
            dict(name="Dwight Schrute", email="dwight.schrute@dundermifflin.com")
        )
        crud_table.insert(
            dict(name="Jim Halpert", email="jim.halpert@dundermifflin.com")
        )
        crud_table.insert(dict(name="Pam Beesly", email="pam.beesly@dundermifflin.com"))

        crud_table.insert(dict(name="Uzumaki Naruto", email="uzumaki.naruto@konoha.jp"))
        crud_table.insert(dict(name="Uchiha Sasuke", email="uchiha.sasuke@konoha.jp"))
        crud_table.insert(dict(name="Haruno Sakura", email="haruno.sakura@konoha.jp"))
        crud_table.insert(dict(name="Hatake Kakashi", email="hatake.kakashi@konoha.jp"))

        self.assertEqual(len(crud_table.select()), 8)
        self.assertEqual(
            len(
                crud_table.select(
                    where="email LIKE :email_like",
                    params=dict(email_like="%@dundermifflin.com"),
                )
            ),
            4,
        )
        self.assertEqual(
            dict(
                crud_table.select(
                    columns=["name"],
                    where="email = :email",
                    params=dict(email="uzumaki.naruto@konoha.jp"),
                )[0]
            ),
            dict(name="Uzumaki Naruto"),
        )

        crud_table.update(
            where="email = :email",
            params=dict(email="uchiha.sasuke@konoha.jp"),
            set=dict(email="uchiha.sasuke@akatsuki.jp"),
        )

        self.assertEqual(
            dict(
                crud_table.select(
                    columns=["email"],
                    where="name = :name",
                    params=dict(name="Uchiha Sasuke"),
                )[0]
            ),
            dict(email="uchiha.sasuke@akatsuki.jp"),
        )

        crud_table.delete(
            where="email = :email", params=dict(email="uchiha.sasuke@akatsuki.jp")
        )

        self.assertEqual(len(crud_table.select()), 7)

        naruto = crud_table.select_one(
            where="email = :email", params=dict(email="uzumaki.naruto@konoha.jp")
        )
        naruto.delete()

        self.assertEqual(len(crud_table.select()), 6)

        sakura = crud_table.select_one(
            where="email = :email", params=dict(email="haruno.sakura@konoha.jp")
        )

        sakura["name"] = "Sakura"
        sakura.save()

        self.assertEqual(
            dict(
                crud_table.select(
                    columns=["name"],
                    where="email = :email",
                    params=dict(email="haruno.sakura@konoha.jp"),
                )[0]
            ),
            dict(name="Sakura"),
        )

    def test_join(self):
        db = get_db()

        consoles = db.create_table()
        consoles = db.update_table(consoles.name, dict(name="consoles"))
        console_name = db.create_column("consoles")
        db.update_column("consoles", console_name.name, dict(name="name"))

        games = db.create_table()
        games = db.update_table(games.name, dict(name="games"))
        game_name = db.create_column("games")
        db.update_column("games", game_name.name, dict(name="name"))
        console_id = db.create_column("games")
        db.update_column("games", console_id.name, dict(name="console_id"))

        ps4 = consoles.insert(
            dict(
                name="Play Station 4",
            )
        )
        nswitch = consoles.insert(dict(name="Nintendo Switch"))

        games.insert(
            [
                dict(name="God of War", console_id=ps4["id"]),
                dict(name="Elden Ring", console_id=ps4["id"]),
                dict(name="Metroid Dread", console_id=nswitch["id"]),
                dict(name="Zelda - Breath of the Wild", console_id=nswitch["id"]),
            ]
        )

        self.assertEqual(len(consoles.select()), 2)
        self.assertEqual(len(games.select()), 4)

        rows = consoles.join(games, on="games.console_id = consoles.id").select(
            columns=["consoles.name", "games.name"]
        )
        self.assertEqual(len(rows), 4)

        self.assertIn(
            {
                "consoles.name": "Nintendo Switch",
                "games.name": "Zelda - Breath of the Wild",
            },
            [dict(row) for row in rows],
        )

    def test_insert_defaults(self):
        db = get_db()

        table = db.create_table()

        self.assertEqual(len(table.select()), 0)
        table.insert()
        self.assertEqual(len(table.select()), 1)
        table.insert({})
        self.assertEqual(len(table.select()), 2)
        table.insert([])
        self.assertEqual(len(table.select()), 2)

    def test_page_response_as_insert_value(self):
        db = get_db()

        table = db.create_table(columns=[dict(name="name"), dict(name="email")])
        page_response = PageResponse(
            data=dict(
                name="Michael Scott",
                email="michael.scott@dundermifflin.com",
            ),
            action="insert",
        )
        table.insert(page_response)

        self.assertEqual(
            [dict(r) for r in table.select(columns=["name", "email"])],
            [dict(name="Michael Scott", email="michael.scott@dundermifflin.com")],
        )



import inspect, flask, werkzeug.exceptions as wz_ex
import abstra_statics
from pathlib import Path

dist_folder = Path(inspect.getfile(abstra_statics)).joinpath("../dist").resolve()


def send_from_dist(filename: str, fallback: str):
    try:
        return flask.send_from_directory(dist_folder, filename)
    except wz_ex.NotFound as e:
        return flask.send_from_directory(dist_folder, fallback)
